package	D2P::Generator::PostgreSQL;
use base 'D2P::Generator';
use D2P::Parser;


sub	strToPgId
{
	my	$string	= shift;

	return $string if $string =~ /^\".*\"$/;

	return "\"$string\"" if (length $string < 32);

	$string =~ s/\_+/_/g;
	return "\"$string\"" if (length $string < 32);

	my	@list	= split /_/, $string;
	my	$old	= $string;
	my	$lsz	= @list;
	my	$rest	= 32 - @list;
	my	$max;

	$string = "";
	foreach my $it (@list)
	{
		$max = $rest / $lsz;
		$max =~ s/\.(.*)$//;
		$max += 1 if ($1 && $1 > 5 && $max < $rest);
		$string .= '_' if $string;

		if	(length $it > $max)
		{
			$string .= substr $it, 0, $max;
			$rest -= $max;
		}
		else
		{
			$string .= $it;
			$rest -= length $string;
		}

		$lsz--;
	}

	print STDERR "WARNING: name \"$old\" changed to \"$string\"\n";
	return "\"$string\"";
}


sub	init
{
	my	$this	= shift;

	$this->SUPER::init();
	$this->{config}->{outfile} = "";
	$this->{config}->{sqlfiles} = [];
}


sub	commandLine
{
	my	$this	= shift;

	return	(
		"-o|--output"	=> {
			desc	=> "Output the PostgreSQL script to the specified file.",
			args	=> ["outfile"],
			uniq	=> 1
		},
		"-s|--sql"	=> {
			desc	=> "Add the contents of the specified file at the end of the script. "
					. "May be repeated for each file to include to the script.",
			argl	=> "sqlfiles",
			args	=> ["sqlfile"],
			uniq	=> 0
		}
	);
}


my $outfile = "";
my @sqlfiles = ();


sub	generate
{
	my	$this = shift;

	$outfile = $this->{config}->{outfile};
	@sqlfiles = map { $_->{sqlfile} } @{$this->{config}->{sqlfiles}};

	return unless $outfile;
	## Prepare output
	if	(-f "$outfile")
	{
		my	($s, @old);

		open(IN, "$outfile") or die "$outfile : $!\n";
		@old = <IN>;
		close(IN);

		open (SQLOLD, ">$outfile~");
		print SQLOLD @old;
		close SQLOLD;

		$s = 0;
		foreach my $l (@old)
		{
			if	(!$s && ($l eq "-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n"))
			{
				$s = 1;
			}
			elsif	($s && ($l eq "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n"))
			{
				last;
			}
			elsif	($s)
			{
				push @customCode, $l;
			}
		}
	}
	open(SQLOUT, ">$outfile");

	&createSQLHeader();
	&createSQLEnums();
	&createSQLTables();
	&copySQLFiles();
	&createSQLFooter();

	close SQLOUT;
}


###
## Create SQL header
#
sub	createSQLHeader
{
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} == 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\n$//;

		my	@lines = split /\n/, $title;
		foreach my $l (@lines)
			{ print SQLOUT "-- $l\n"; }
		print SQLOUT "\n";
		last;
	}

	print SQLOUT "-- Input file  : $infile\n";
	print SQLOUT "-- Output file : $outfile\n" if $outfile;
	print SQLOUT "\n";
	print SQLOUT "-- Defines the following modules : \n";
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;

		print SQLOUT "--  $title\n";
	}
	print SQLOUT "\n";

	my	@requiredModules = ();
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'pack' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;
		push @requiredModules, $title;
	}
	if	(@requiredModules)
	{
		print SQLOUT "-- Requires the following modules :\n--  ";
		print SQLOUT join ("\n--  ", @requiredModules);
		print SQLOUT "\n\n";
	}
	print SQLOUT "BEGIN TRANSACTION;\n\n\n";
}


###
## Create SQL tables for enumerations according to data created by checkAndModifyInput()
#
sub	createSQLEnums
{
	foreach my $enum (@enumerations)
	{
		my	$l = 0;
		foreach my $w1 (@{$enum->{contents}})
		{
			my	$w1l = length($w1);
			($w1l > $l) && ($l = $w1l);
		}

		print SQLOUT "-- Enumeration table for " . $enum->{fieldname} . "\n\n";
		print SQLOUT "CREATE TABLE " . $enum->{tablename} . "(\n\tid\tSERIAL PRIMARY KEY,\n\t";
		print SQLOUT "txt\tVARCHAR(" . ($l+1) . ") NOT NULL\n);\n";
		foreach my $w2 (@{$enum->{contents}})
		{
			print SQLOUT "INSERT INTO ". $enum->{tablename} . "(txt) VALUES('$w2');\n";
		}
		(my $tnVal = $enum->{tablename}) =~ s/\"//g;
		print SQLOUT "\nINSERT INTO d2p_lst_list(d2p_name, dia_name) VALUES ('$tnVal', '";
		print SQLOUT $enum->{compname} . "');\n";
		print SQLOUT "\n\n";
	}
}


###
## Create SQL tables for each included module
#
sub	createSQLTables
{
	foreach my $gro (sort keys %groupsOrdered)
	{
		foreach my $g (@{$groupsOrdered{$gro}})
		{
			next unless ($g->{type} eq 'box' && @{$g->{classes}});

			my	$title = $g->{topLevelText};
			$title =~ s/\n/\n-- /g;
			print SQLOUT "-- Tables for $title\n\n";
			foreach my $o (@{$g->{classes}})
			{
				my	$x = 0;
				my	@pkey = ();
				my	@references = ();
				my	$tnam = strToPgId($o->{name});

				print SQLOUT "CREATE TABLE $tnam (";
				foreach my $f (@{$o->{fields}})
				{
					my	$fnam = strToPgId($f->{name});
					$x ? (print SQLOUT ",") : ($x=1);
					print SQLOUT "\n\t$fnam";
					my	$l = 56 - length $fnam;
					if	($l <= 0)
					{
						print SQLOUT " ";
					}
					else
					{
						my	$nt = $l / 8;
						$nt++ if $nt =~ s/\..*$//;
						print SQLOUT "\t" x $nt;
					}
					print SQLOUT $f->{type};

					if	($f->{not_null})
					{
						print SQLOUT " NOT NULL";
					}
					if	($f->{references})
					{
						# FIXME : has to be an option
						#my	($table, $field) = split /\./, $f->{references};
						#print SQLOUT " REFERENCES $table";
						#$field && print SQLOUT "($field)";
						#print SQLOUT " ON DELETE CASCADE";
						my	($table, $field) = map { strToPgId($_); } (split (/\./, $f->{references}));
						my	$tmpStr = "'$tnam', '$fnam', '$table', '$field', ";

						if	($f->{refCascade} & 1)
							{ $tmpStr .= "true"; }
						else
							{ $tmpStr .= "false"; }
						$tmpStr .= ", ";
						if	($f->{refCascade} & 2)
							{ $tmpStr .= "true"; }
						else
							{ $tmpStr .= "false"; }
						push @references, $tmpStr;
					}
				}
				if	($o->{pkey})
				{
					my	@opk = split /,/, $o->{pkey};
					@opk = map { strToPgId($_) } (@opk);
					print SQLOUT ",\n\tPRIMARY KEY(" . join(",", @opk) . ")";
				}
				print SQLOUT "\n)";

				if	(@{$o->{inherits}})
				{
					my	$y = 0;

					print SQLOUT " INHERITS (";
					foreach my $super (@{$o->{inherits}})
					{
						$y ? (print SQLOUT ", ") : ($y = 1);
						print SQLOUT strToPgId($super->{name});
					}
					print SQLOUT ")";
				}

				print SQLOUT ";\n";
				$tnam =~ s/\"//g;
				print SQLOUT "INSERT INTO d2p_lst_list(d2p_name, dia_name) VALUES ('$tnam', '$o->{name}');\n";
				foreach my $i (@{$o->{indices}})
				{
					print SQLOUT "CREATE " . ($i->{unique} ? 'UNIQUE ' : '') . "INDEX ";
					print SQLOUT strToPgId($i->{name}) . " ON " . strToPgId($o->{name});
					$i->{type} && print SQLOUT " USING " . $i->{type};
					print SQLOUT " (" . join(', ', map { strToPgId($_); } (@{$i->{fields}})) . ");\n";
				}
				foreach my $ref (@references)
				{
					$ref =~ s/\"//g;
					print SQLOUT "INSERT INTO d2p_ref_list(src_table, src_field, dest_table, dest_field, del_cascade, upd_cascade)\n";
					print SQLOUT "VALUES ($ref);\n";
				}
				print SQLOUT "\n";
			}
			print SQLOUT "\n";
		}
	}
}


###
## Copy additional SQL files
#
sub	copySQLFiles
{
	return unless @sqlfiles;

	print SQLOUT "\n\n-- \n-- ADDITIONAL SQL FILES\n-- \n";
	foreach my $f (@sqlfiles)
	{
		unless	(open(SFIC, $f))
		{
			next;
		}
		print SQLOUT "\n-- \n -- SQL code from $f\n-- \n\n";
		my @cff;
		@cff = <SFIC>;
		close SFIC;
		print SQLOUT @cff;
	}
}


###
## Create SQL footer
#
sub	createSQLFooter
{
	print SQLOUT "\n\n-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n";
	if	(@customCode)
	{
		print SQLOUT @customCode;
	}
	else
	{
		print SQLOUT "\n" x 4;
	}
	print SQLOUT "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n\n";
	print SQLOUT "SELECT f_d2p_inh_update();\nSELECT f_d2p_ref_update();\n\n";
	print SQLOUT "COMMIT TRANSACTION;\n";
	print SQLOUT "-- SQL code generated by dia2postgres.pl\n";
	print SQLOUT "-- $VERSION\n"; 
	print SQLOUT "-- on " . `date`;
}




1;
