#!/usr/bin/perl

use	XML::Parser;
use	D2P::Parser;
use	D2P::Generator;

BEGIN {
	# Set version
	my $cvsId = '$Id: dia2postgres.pl,v 1.23 2002/11/17 16:29:50 uid5000 Exp $';
	$cvsId =~ s/^\$Id: //;
	$cvsId =~ s/\$$//;
	$main::VERSION	= $cvsId;
}

# Activate debugging by setting this variable to 1
$DEBUG		= 0;

# Generator modules to load
@GENMODS	= ("PostgreSQL", "PHPClasses", "PgDrop");


###                                     ###
#### INPUT VALIDATION AND MODIFICATION ####
###                                     ###

###
## This function creates a valid PostgreSQL identifier from a string
#
sub	strToPgId
{
	my	$string	= shift;

	return $string if $string =~ /^\".*\"$/;

	return "\"$string\"" if (length $string < 32);

	$string =~ s/\_+/_/g;
	return "\"$string\"" if (length $string < 32);

	my	@list	= split /_/, $string;
	my	$old	= $string;
	my	$lsz	= @list;
	my	$rest	= 32 - @list;
	my	$max;

	$string = "";
	foreach my $it (@list)
	{
		$max = $rest / $lsz;
		$max =~ s/\.(.*)$//;
		$max += 1 if ($1 && $1 > 5 && $max < $rest);
		$string .= '_' if $string;

		if	(length $it > $max)
		{
			$string .= substr $it, 0, $max;
			$rest -= $max;
		}
		else
		{
			$string .= $it;
			$rest -= length $string;
		}

		$lsz--;
	}

	print STDERR "WARNING: name \"$old\" changed to \"$string\"\n";
	return "\"$string\"";
}


###
## Validate and modify the whole structure
#
sub	checkAndModifyInput
{
	foreach my $o (@allObjects)
	{
		next unless $o;

		# Handle local fields
		if	(	$o->{to} eq 'field'
			&&	$o->{parent} && $o->{parent}->{parent}
			&&	$o->{parent}->{parent}->{type} eq 'box'
			)
		{
			# Handle enumerations
			if	($o->{type} =~ /^\{ .* \}$/ && !$o->{references})
			{
				my	$t = $o->{type};
				my	@n;
				my	$dsc = {};

				$t =~ s/\{ (.*) \}$/$1/;
				$t =~ s/\s+//g;
				@n = split /,/, $t;
				$t = $o->{parent}->{name} . "." . $o->{name};
				$dsc->{fieldname} = $t;
				$t =~ s/\./_/g;
				$dsc->{compname} = "e_$t";
				$dsc->{tablename} = strToPgId("e_$t");
				$dsc->{contents} = [@n];
				push @enumerations, $dsc;

				$o->{type} = "INT";
				$o->{references} = $dsc->{tablename} . ".id";
				$o->{refCascade} = 3;
			}
		}
		# Sort classes inside groups
		elsif	($o->{to} eq 'group' && $o->{type} eq 'box')
		{
			my $nc = [];
			@$nc = sort { $a->{order} <=> $b->{order} } @{$o->{classes}};
			$o->{classes} = $nc;
		}
	}

	my	@tmp	= grep { $_->{type} == 4 } values %lines;
	foreach my $ir (@tmp)
	{
		my	$target = $classes{$ir->{from}->{class}};
		next unless defined $target;
		my	$super = $classes{$ir->{to}->{class}};
		next unless defined $super;
		push @{$target->{inherits}}, $super;
	}

	@tmp = grep { defined($_->{type}) && ($_->{type} != 4) } values %lines;
	foreach my $rel (@tmp)
	{
		next if $rel->{from}->{dest} < 8;

		my	$target	= $classes{$rel->{from}->{class}};
		next unless defined $target;
		my	$targetFieldNb = $rel->{from}->{dest} - 8;
		$targetFieldNb -= $targetFieldNb % 2;
		$targetFieldNb /= 2;
		next unless $targetFieldNb < @{$target->{fields}};
		my	$targetField = $target->{fields}->[$targetFieldNb];
		next if (	$targetField->{references}
			||	!$targetField->{mode}
			);

		my	$ref	= $classes{$rel->{to}->{class}};
		next unless defined $ref;
		my	$refField = '';

		if	($rel->{to}->{dest} < 8)
		{
			next unless	(	@{$ref->{inherits}}
					||	($ref->{pkey} && $ref->{pkey} !~ /,/)
					);
			if	($ref->{pkey} && $ref->{pkey} !~ /,/)
			{
				$refField = (grep { $_->{name} eq $ref->{pkey} } @{$ref->{fields}})[0];
			}
			elsif	(@{$ref->{inherits}})
			{
				# WARNING : this will loop forever with broken, recurrent inheritance
				my ($tmp, @checklist) = (0);
				push @checklist, @{$ref->{inherits}};
				while (!$refField and $tmp = shift @checklist)
				{
					if	($tmp->{pkey} && $tmp->{pkey} !~ /,/)
						{ $refField = (grep { $_->{name} eq $tmp->{pkey} } @{$tmp->{fields}})[0]; }
					elsif	(@{$tmp->{inherits}})
						{ push @checklist, @{$tmp->{inherits}}; }
				}
			}
		}
		else
		{
			my	$refFieldNb = $rel->{to}->{dest} - 8;
			$refFieldNb -= $refFieldNb % 2;
			$refFieldNb /= 2;
			next unless $refFieldNb < @{$ref->{fields}};
			$refField = $ref->{fields}->[$refFieldNb];
			next if ($refField->{mode} != 1);
		}
		next unless $refField;

		$targetField->{references} = $ref->{name} . '.' . $refField->{name};
		$targetField->{refCascade} = $rel->{type};
	}

	$tmpFunc = sub {
		my $class = shift;
		my $level = shift;
		$DEBUG && print STDERR (" " x $level) . "$class->{name} class : finding inheritance\n";
		$DEBUG && print STDERR (" " x $level) . " > $class->{name} : primary key = $class->{pkey}\n";
		foreach my $superClass (@{$class->{inherits}})
		{
			$DEBUG && print STDERR (" " x $level) . "  > inherits $superClass->{name}\n";
			&$tmpFunc($superClass, $level + 1) unless $superClass->{indone};
			$class->{pkey} .= ($class->{pkey} ? "," : "") . $superClass->{pkey}
				if	$superClass->{pkey};
		}
		$DEBUG && print STDERR (" " x $level) . " > $class->{name} : primary key = $class->{pkey}\n";
		$class->{indone} = 1;
	};
	foreach my $cn (values %classes)
	{
		&$tmpFunc($cn, 0) unless $cn->{indone};
	}
	undef $tmpFunc;
}



###                    ###
#### HELPER FUNCTIONS ####
###                    ###


###
## Parses command line arguments according to a description
#
sub	parseArguments
{
	my	$dptr	= shift @_;
	my	%desc	= %{$dptr};
	my	@args	= @_;
	my	$state	= 0;
	my	%opts	= ();
	my	%curarg	= ();
	my	$clel;
	my	$argc;

	return unless @args;

	$DEBUG && print STDERR "Command line arguments : '" . join("'    '", @args) . "'\n";

	foreach my $k (keys %desc)
	{
		my	@l = split /\|/, $k;
		foreach my $opt (@l)
		{
			$opts{$opt} = $k;
		}
	}

	foreach my $arg (@args)
	{
		$DEBUG && print STDERR "arg = $arg   state = $state\n";
		if	($state == 0)
		{
			if	(!grep(/^$arg$/, keys %opts))
			{
				die "Invalid argument : $arg\n";
			}

			my $k = $opts{$arg};
			%curarg = %{$desc{$k}};
			$DEBUG && print STDERR "curarg{uniq} = $curarg{uniq}    curarg{found} = $curarg{found}\n";
			die "Syntax error : $arg found twice in command line\n"
				if	($curarg{uniq} && $curarg{found});

			$desc{$k}->{found} = 1;

			if	($curarg{$instance})
			{
				if	($curarg{bvar})
				{
					my $cmd = "\$curarg{instance}->{config}->{$curarg{bvar}} = !";
					$cmd .= "\$curarg{instance}->{config}->{$curarg{bvar}};";
					eval $cmd;
				}
				elsif	($curarg{args} && !$curarg{argl})
				{
					$argc = 0;
					$state = 1;
				}
				elsif   ($curarg{args})
				{
					$argc = 0;
					$clel = {};
					$state = 2;
				}
				elsif	($curarg{hdlr})
				{
					my $cmd = "\$curarg{instance}->$curarg{hdlr} (\$dptr);";
					eval $cmd;
					if	($@)
					{
						(my $err = $@) =~ s/\n/ /g;
						print STDERR "WARNING : handler for argument $arg failed ($err)\n";
					}
				}
				else
				{
					print STDERR "WARNING : $arg is not implemented\n";
				}
			}
			else
			{
				if	($curarg{bvar})
				{
					my $cmd = "\$$curarg{bvar} = !$curarg{bvar};";
					eval $cmd;
				}
				elsif	($curarg{args} && !$curarg{argl})
				{
					$argc = 0;
					$state = 1;
				}
				elsif   ($curarg{args})
				{
					$argc = 0;
					$clel = {};
					$state = 2;
				}
				elsif	($curarg{hdlr})
				{
					&{$curarg{hdlr}} ($dptr);
				}
				else
				{
					print STDERR "Warning : $arg is not implemented\n";
				}
			}
		}
		elsif	($state == 1)
		{
			# Standard argument
			die "Internal error in command line parsing\n"
				if	(!$curarg{args}->[$argc]);

			my $cmd;
			
			if	($curarg{instance})
			{
				$cmd = "\$curarg{instance}->{config}->{" . $curarg{args}->[$argc] . "}";
			}
			else
			{
				$cmd = "\$" . $curarg{args}->[$argc];
			}
			$cmd .= " = \$arg;";
			eval $cmd;

			$argc++;
			if ($argc == @{$curarg{args}})
			{
				if	($curarg{hdlr})
				{
					if	($curarg{instance})
					{
						my $cmd = "\$curarg{instance}->$curarg{hdlr} (\$dptr);";
						eval $cmd;
						if	($@)
						{
							(my $err = $@) =~ s/\n/ /g;
							print STDERR "WARNING : handler for argument $arg failed ($err)\n";
						}
					}
					else
					{
						&{$curarg{hdlr}} ($dptr);
					}
				}
				$state = 0;
			}
		}
		elsif	($state == 2)
		{
			# List argument
			die "Internal error in command line parsing\n"
				if	(!$curarg{args}->[$argc]);

			my $cmd = "\$clel->{" . $curarg{args}->[$argc] . "} = \$arg;";
			eval $cmd;

			$argc++;
			if	($argc == @{$curarg{args}})
			{
				&{$curarg{hdlr}} ($dptr) if $curarg{hdlr};
				if	($curarg{instance})
				{
					eval "push \@{\$curarg{instance}->{config}->{$curarg{argl}}}, \$clel;";
					if	($curarg{hdlr})
					{
						my $cmd = "\$curarg{instance}->$curarg{hdlr} (\$dptr);";
						eval $cmd;
						if	($@)
						{
							(my $err = $@) =~ s/\n/ /g;
							print STDERR "WARNING : handler for argument $arg failed ($err)\n";
						}
					}
				}
				else
				{
					eval "push \@$curarg{argl}, \$clel;";
					if	($curarg{hdlr})
					{
						&{$curarg{hdlr}} ($dptr);
					}
				}
				$state = 0;
			}
		}
	}

	die "Error : unexpected end of command line\n"
		if	($state != 0);
}


###
## Shows command line help
#
sub		usage
{
	my	$dptr	= shift @_;
	my	%desc	= %{$dptr};
	my	$maxlen	= 0;
	my	$maxCol = $ENV{"COLUMNS"} ? $ENV{"COLUMNS"} : 80;

	$DEBUG && print STDERR "maxCol = $maxCol " . $ENV{COLUMNS} . "\n";

	print STDERR "dia2postgres - A Perl script that converts a Dia diagram into\n";
	print STDERR "               a PostgreSQL script, and maybe more.\n\n";
	print STDERR "$VERSION\n";
	print STDERR "(c) 2002 by Emmanuel Benoit <tseeker\@nocternity.net>\n\n";

	foreach my $a (keys %desc)
		{ $maxlen = length($a) if (length($a) > $maxlen); }
	$maxlen += 3;
	exit(1) if ($maxlen >= $maxCol - 15);

	print STDERR "Usage : $0 [<arguments>]\n\n";
	print STDERR "Where arguments are :\n";

	foreach my $arg (sort keys %desc)
	{
		my	@d	= split / /, $desc{$arg}->{desc};
		my	$col	= $maxlen + 1;
		my	$w;

		print STDERR "\n";
		print STDERR " " . $arg . (" "x($maxlen - length($arg)));
		while	($w = shift @d)
		{
			if	($col + length($w) + 1 >= $maxCol)
			{
				$col = $maxlen + 4;
				print STDERR "\n" . (" "x$col);
			}
			print STDERR "$w ";
			$col += length($w) + 1;
		}
		print STDERR "\n";
		print STDERR (" "x($maxlen + 1)) . "USAGE : $arg <" . join("> <", @{$desc{$arg}->{args}}) . ">\n"
			if @{$desc{$arg}->{args}};
	}
	print "\n\n";

	exit(1);
}


###             ###
#### MAIN CODE ####
###             ###

###
## Command line arguments definitions and default values
#

%argsDescription = (
	"-i|--input"	=> {
		desc	=> "Read the Dia XML formatted input from the specified file instead of stdin.",
		args	=> ["infile"],
		uniq	=> 1
	},
	"-h|--help"	=> {
		desc	=> "Display this message.",
		uniq	=> 1,
		hdlr	=> \&usage
	}
);

$infile = "";

## Load modules
@genInsts = ();
foreach my $module (@GENMODS)
{
	my	$instance = undef;
	eval "use D2P::Generator::$module;";
	if	($@)
	{
		my $err = $@;
		$err =~ s/\n/ /g;
		print STDERR "WARNING : could not load module $module ($err)\n";
		next;
	}
	$instance = "D2P::Generator::$module"->new();
	if	(defined($instance))
	{
		push @genInsts, $instance;
	}
}

## Parse command line
foreach my $instance (@genInsts)
{
	## Collect command line arguments from modules
	my %args = $instance->commandLine();
	foreach my $arg (keys %args)
	{
		my $a = $arg;
		$a =~ s/([^A-Za-z0-9])/\\$1/g;
		next if grep (/^$a$/, keys %argsDescription);
		$argsDescription{$arg} = $args{$arg};
		$argsDescription{$arg}->{instance} = $instance;
	}
}
&parseArguments(\%argsDescription, @ARGV);

## Read input
@realin = ();
if	($infile ne "")
{
	open(FIC, $infile) or die "Error : could not open $infile\n";
	@realin = <FIC>;
	close(FIC);
}
else
{
	@realin = <STDIN>;
}

## Init xml parser and parse input file
$parser = new XML::Parser(Style => 'Stream', Pkg => D2P::Parser);
$parser->parse(join("", @realin));
&checkAndModifyInput();

## Generate code
foreach my $inst (@genInsts)
{
	$inst->generate();
}

exit;



